/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package net.sourceforge.basher;

import java.util.Iterator;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.atomic.AtomicBoolean;

import bsh.EvalError;
import bsh.Interpreter;
import org.ops4j.gaderian.Registry;
import org.ops4j.gaderian.impl.RegistryBuilder;

/**
 * Main entry point to the Basher when using the command line mode or for applications not using Gaderian.
 *
 * @author Johan Lindquist
 * @version 1.0
 */
public class Basher
{
    /**
     * Basher Gaderian registry instance
     */
    private static Registry _registry;

    /**
     * Flag indicating whether or not a shutdown has been initiated
     */
    private static AtomicBoolean _shutdownInitiated = new AtomicBoolean(false);

    /**
     * Sets the registry this Basher instance should use.
     *
     * @param registry The Gaderian registry to use.
     * @throws IllegalArgumentException If the registry has already been set.
     */
    public static void setRegistry(final Registry registry)
    {
        if (_registry != null)
        {
            throw new IllegalStateException("Basher has already been initialized");
        }
        _registry = registry;
    }

    /**
     * Retrieves the registry used by Basher
     *
     * @return The registry in use
     * @throws IllegalStateException If the registry has not been set.
     */
    public static Registry getRegistry()
    {
        if (_registry == null)
        {
            throw new IllegalStateException("Basher has not been initialized");
        }
        if (_shutdownInitiated.get())
        {
            throw new IllegalStateException("Basher shutdown has been initiated");
        }
        return _registry;
    }

    /**
     * Convenience method to be used when a thread cleanup of the registry should happen.
     */
    public static void fireCleanUpThread()
    {
        if (_shutdownInitiated.get())
        {
            return;
        }
        try
        {
            if (_registry != null)
            {
                _registry.cleanupThread();
            }
        }
        catch (Throwable e)
        {
            e.printStackTrace();
        }
    }

    /**
     * Retrieves a service from the Basher (Gaderian) registry.  This is only to be used by
     * application that do not themselves use Gaderian.
     *
     * @param service The class of the service to retrieve.
     * @return An instance of the service.
     * @throws Exception If the registry fails to intialize or the service does not exist.
     */
    public static <T> T getService(final Class<T> service) throws Exception
    {
        if (_registry == null)
        {
            _registry = RegistryBuilder.constructDefaultRegistry();
        }
        return (T)_registry.getService(service);
    }

}
