package net.sourceforge.basher;

import java.util.ArrayList;
import java.util.List;

/**
 * Container class for holding task configuration options and manipulating certain aspects of it.
 *
 * @author Johan Lindquist
 */
public class TaskConfiguration
{
    /**
     * Instance weight as a number between 0 and 100 - 0 indicating the task would not run, 100 that it always runs.
     * Defaults to 100 (always runs).
     */
    private int _weight = 100;

    /**
     * instance mirror of weight as the "real" float.  This is to avoid stomping on the same number.
     */
    private double _doubleWeight = _weight;

    /**
     * Instance backup of the weight - kept for reset
     */
    private int _originalWeight = _weight;

    /**
     * instance inertia
     */
    private double _inertia = 1.0d;

    /**
     * maximum invocation - if non-zero, this task should execute at most _maxInvocations times
     */
    private int _maxInvocations = 0;

    /**
     * Defines the time (internal Basher time) at which point the task will start running.  Defaults to 0
     * (start initially).
     */
    private long _runFrom = 0;

    /**
     * Defines the time (internal Basher time) at which point the task will stop running.  Defaults to
     * <code>-1</code> (run forever).
     */
    private long _stopAfter = Long.MAX_VALUE;

    /**
     * Defines the maximum amount of time allowed for a task to run.  If the task takes longer than the specified
     * time, the invocation will be considered a failure.  Defaults to 0 (no max time)
     */
    private int _maxTime = 0;

    /**
     * Defines the registered task name
     */
    private String _taskName;

    /**
     * Defines the phases for which the task should be available in.  By default, a task is always available in the
     * <code>RUN</code> phase.
     */
    private List<Phase> _applicablePhases = new ArrayList<Phase>();

    public TaskConfiguration()
    {
        _applicablePhases.add( Phase.RUN );
    }

    /**
     * Retrieves the weight of this <code>AbstractTask</code>.  This could be used by certain schedulers to allow for
     * more fine grained control of task scheduling
     *
     * @return A value between 1 and 100 (higher signals a wish to run more frequently)
     */
    public Integer getWeight()
    {
        return _weight;
    }

    /**
     * Retrieves the inertia of this <code>TaskConfiguration</code>.
     * See {@link net.sourceforge.basher.TaskConfiguration#setInertia(Double)}.
     *
     * @return The rate of change for this instance.
     */
    public Double getInertia()
    {
        return _inertia;
    }

    /**
     * Sets the inertia of this <code>AbstractTask</code>.  This is used to control the change of
     * the weight assocaited with this task, calculated for each invocation of the task (if it is
     * not 1.0).
     *
     * @param inertia The inertia to give this instance.  This should be a value between 0.1 and 2.0.
     */
    public void setInertia( final Double inertia )
    {
        _inertia = inertia;
    }

    /**
     * Resets the weight calculattions of this <code>TaskConfiguration</code> instance.
     */
    public void reset()
    {
        _weight = _originalWeight;
        _doubleWeight = _weight;
    }

    /**
     * Sets the weight for the task.  This <i>could</i> be used by a scheduler to control the invocations of the
     * <code>Task</code>.
     *
     * @param weight The weight of the task.  Should be a value between 1 and 100.
     */
    public void setWeight( final Integer weight )
    {
        _weight = weight;
        // Mirror the weight in the float calculations too ...
        _doubleWeight = _weight;

        // And back these settings up ...
        _originalWeight = weight;
        _doubleWeight = weight;

    }

    /**
     * Sets the maximum number of times the task should be invoked.  See
     * {@link net.sourceforge.basher.TaskConfiguration#getMaxInvocations()} for more information.
     *
     * @param maxInvocations The number of invocations to do.  A value of 0 (zero) or less indicates that the task
     *                       should always be invoked.
     */
    public void setMaxInvocations( final Integer maxInvocations )
    {
        _maxInvocations = maxInvocations;
    }

    /**
     * Retrieves the maximum number of times the task should be invoked.  The number returned <i>could</i> be used
     * by a scheduler to control the invocation of the <code>Task</code>.
     *
     * @return The maximum number of invocations
     */
    public Integer getMaxInvocations()
    {
        return _maxInvocations;
    }


    /**
     * Recalculates the weight of this <code>Task</code>.  This will take the inertia and multiply it with the current
     * weight and cast the result to an integer.
     */
    public void reCalculateWeight()
    {
        // Only do anything if
        if ( Double.compare( _inertia, 1.0D ) != 0 )
        {
            // This ensure we dont stand stompin' on the same number
            // if the inertia is small ...
            _doubleWeight = _doubleWeight * _inertia;
            _weight = ( int ) _doubleWeight;
        }
    }

    /**
     * Retrieves the default time from which this task should run.
     *
     * @return The time from which this task should run.  0 indicates run from start, otherwise specifies the seconds from
     *         start of the run phase.
     */
    public Long getRunFrom()
    {
        return _runFrom;
    }

    /**
     * Retrieves the default time from which this task should stop running.
     *
     * @return The time from which this task should stop running.  Always returns {@link java.lang.Long#MAX_VALUE}
     *         (run forever) unless overridden by sub-class.
     */
    public Long getStopAfter()
    {
        return _stopAfter;
    }

    /**
     * Sets the time from which this task should run.
     *
     * @param runFrom The time from which this task should run.  0 indicates run from start, otherwise specifies the seconds from
     *                start of the run phase.
     */
    public void setRunFrom( final Long runFrom )
    {
        _runFrom = runFrom;
    }

    /**
     * Sets the default time from which this task should stop running.
     *
     * @param stopAfter The time from which this task should stop running.  Always returns {@link java.lang.Long#MAX_VALUE}
     *                  (run forever) unless overridden by sub-class.
     */
    public void setStopAfter( final Long stopAfter )
    {
        _stopAfter = stopAfter;
    }


    /**
     * Clears the list of applicable phases.
     */
    public void clearApplicablePhases()
    {
        _applicablePhases.clear();
    }

    /**
     * Adds the specified phase to the set of applicable phases.  This method will ensure there are no duplicate entries.
     *
     * @param phase The phases to add
     */
    public void addApplicablePhase( Phase phase )
    {
        if ( phase == null )
        {
            throw new NullPointerException( "phase" );
        }
        if ( !_applicablePhases.contains( phase ) )
        {
            _applicablePhases.add( phase );
        }
    }

    /**
     * Retrieves the <code>Phase</code>s in which this <code>Task</code> is run.  This, by default, returns only
     * <code>Phase.RUN</code>.
     *
     * @return List of phases this <code>Task</code> should run in - only RUN phase is included by default
     */
    public List<Phase> getApplicablePhases()
    {
        return _applicablePhases;
    }

    /**
     * Returns the default max time for <code>Task</code>s.
     *
     * @return Always returns -1, signifying no maximum time
     */
    public Integer getMaxTime()
    {
        return _maxTime;
    }

    public void setMaxTime( final Integer maxTime )
    {
        _maxTime = maxTime;
    }

    /**
     * Retrieves the name that the <code>Task</code> instance should be known as.
     *
     * @return The name of the task instance.
     */
    public String getTaskName()
    {
        return _taskName;
    }

    /**
     * Sets the name that the <code>Task</code> instance should be known as.
     *
     * @param taskName The name of the instance.
     */
    public void setTaskName( final String taskName )
    {
        _taskName = taskName;
    }

    /**
     * @param taskConfiguration
     */
    public void initialize( final TaskConfiguration taskConfiguration )
    {
        _applicablePhases = new ArrayList<Phase>( taskConfiguration.getApplicablePhases() );
        _inertia = taskConfiguration.getInertia();
        _maxInvocations = taskConfiguration.getMaxInvocations();
        _maxTime = taskConfiguration.getMaxTime();
        _runFrom = taskConfiguration.getRunFrom();
        _stopAfter = taskConfiguration.getStopAfter();
        _taskName = taskConfiguration.getTaskName();
        _weight = taskConfiguration.getWeight();
    }


    @Override
    public boolean equals( final Object o )
    {
        if ( this == o )
        {
            return true;
        }
        if ( !( o instanceof TaskConfiguration ) )
        {
            return false;
        }

        final TaskConfiguration that = ( TaskConfiguration ) o;

        if ( Double.compare( that._doubleWeight, _doubleWeight ) != 0 )
        {
            return false;
        }
        if ( Double.compare( that._inertia, _inertia ) != 0 )
        {
            return false;
        }
        if ( _maxInvocations != that._maxInvocations )
        {
            return false;
        }
        if ( _maxTime != that._maxTime )
        {
            return false;
        }
        if ( _originalWeight != that._originalWeight )
        {
            return false;
        }
        if ( _runFrom != that._runFrom )
        {
            return false;
        }
        if ( _stopAfter != that._stopAfter )
        {
            return false;
        }
        if ( _weight != that._weight )
        {
            return false;
        }
        if ( _taskName != null ? !_taskName.equals( that._taskName ) : that._taskName != null )
        {
            return false;
        }

        return true;
    }

    @Override
    public int hashCode()
    {
        int result;
        long temp;
        result = _weight;
        temp = _doubleWeight != +0.0d ? Double.doubleToLongBits( _doubleWeight ) : 0L;
        result = 31 * result + ( int ) ( temp ^ ( temp >>> 32 ) );
        result = 31 * result + _originalWeight;
        temp = _inertia != +0.0d ? Double.doubleToLongBits( _inertia ) : 0L;
        result = 31 * result + ( int ) ( temp ^ ( temp >>> 32 ) );
        result = 31 * result + _maxInvocations;
        result = 31 * result + ( int ) ( _runFrom ^ ( _runFrom >>> 32 ) );
        result = 31 * result + ( int ) ( _stopAfter ^ ( _stopAfter >>> 32 ) );
        result = 31 * result + _maxTime;
        result = 31 * result + ( _taskName != null ? _taskName.hashCode() : 0 );
        result = 31 * result + ( _applicablePhases != null ? _applicablePhases.hashCode() : 0 );
        return result;
    }

    @Override
    public String toString()
    {
        return "TaskConfiguration{" +
                "_weight=" + _weight +
                ", _doubleWeight=" + _doubleWeight +
                ", _originalWeight=" + _originalWeight +
                ", _inertia=" + _inertia +
                ", _maxInvocations=" + _maxInvocations +
                ", _runFrom=" + _runFrom +
                ", _stopAfter=" + _stopAfter +
                ", _maxTime=" + _maxTime +
                ", _taskName='" + _taskName + '\'' +
                '}';
    }
}