/*
 *
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * 	http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sourceforge.basher.impl;

import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.locks.*;

import net.sourceforge.basher.*;
import net.sourceforge.basher.events.*;
import net.sourceforge.basher.internal.*;
import org.apache.commons.logging.Log;

/**
 * @author Johan Lindquist
 * @version 1.0
 */
public class TaskManagerImpl implements TaskManager, BasherEventListener
{
    private List<TaskProvider> _taskProviders;
    private Set<TaskContribution> _taskContributions;
    private List<TaskExecutionContext> _availableTaskExecutionContexts;
    private Randomizer _randomizer;
    private Log _logger;
    private EventManager _eventManager;
    private TaskCreator _taskCreator;
    private TaskDecorator _taskDecorator;
    private Lock _taskLock = new ReentrantLock();
    private List<ModuleDescriptorTaskContribution> _moduleDescriptorTaskContributions;
    private TaskConfigurationExtractor _taskConfigurationExtractor;

    public void setTaskConfigurationExtractor(final TaskConfigurationExtractor taskConfigurationExtractor)
    {
        _taskConfigurationExtractor = taskConfigurationExtractor;
    }

    public void setTaskCreator(final TaskCreator taskCreator)
    {
        _taskCreator = taskCreator;
    }

    public void setTaskDecorator(final TaskDecorator taskDecorator)
    {
        _taskDecorator = taskDecorator;
    }

    public void setEventManager(final EventManager eventManager)
    {
        _eventManager = eventManager;
    }

    public void setLog(final Log logger)
    {
        _logger = logger;
    }

    public void setRandomizer(final Randomizer randomizer)
    {
        _randomizer = randomizer;
    }

    public void setTaskProviders(final List<TaskProvider> taskProviders)
    {
        _taskProviders = taskProviders;
    }

    public void addTask(final TaskContribution taskContribution)
    {
        checkInitialized();
        if (taskContribution == null)
        {
            throw new NullPointerException( "taskContribution" );
        }
        _taskContributions.add(taskContribution);

        // Signal the fact
        _eventManager.publish(new TasksAvailableEvent());

    }

    public TaskContribution addTask( final Class taskClass, final TaskContribution... followers)
    {
        if (taskClass == null)
        {
            throw new NullPointerException( "taskClass" );
        }
        
        // Create a task configuration for the task (using the annotations)
        final TaskConfiguration taskConfiguration = _taskConfigurationExtractor.extractTaskConfiguration(taskClass);

        // Create an executable <code>Task</code> from the class name
        final Task task = _taskCreator.createTask(taskClass);

        // Create a task contribution of the task configuration and the task
        final TaskContribution taskContribution = new TaskContribution(task, taskConfiguration);

        // Add any followers
        taskContribution.setFollowers(Arrays.asList(followers));

        addTask(taskContribution);

        return taskContribution;
    }

    public TaskContribution addTask( final Object taskInstance, final TaskContribution... followers)
    {
        if (taskInstance == null)
        {
            throw new NullPointerException( "taskInstance" );
        }

        // Create a task configuration for the task (using the annotations)
        final TaskConfiguration taskConfiguration = _taskConfigurationExtractor.extractTaskConfiguration(taskInstance.getClass());

        final Task task = _taskDecorator.decorateInstance(taskInstance);

        // Create a task contribution of the task configuration and the task
        final TaskContribution taskContribution = new TaskContribution(task, taskConfiguration);

        // Add any followers
        taskContribution.setFollowers(Arrays.asList(followers));

        addTask(taskContribution);

        return taskContribution;
    }

    public TaskExecutionContext getNextTaskExecutionContext()
    {
        checkInitialized();
        _taskLock.lock();
        try
        {
            if (_availableTaskExecutionContexts.size() == 0)
            {
                _logger.debug("No active tasks available");
                return null;
            }

            final int nextRandom = _randomizer.getRandomInt(_availableTaskExecutionContexts.size());
            final TaskExecutionContext taskExecutionContext = _availableTaskExecutionContexts.get(nextRandom);
            if (taskExecutionContext == null)
            {
                return null;
            }
            return taskExecutionContext;
        }
        finally
        {
            _taskLock.unlock();
        }
    }

    public Set<TaskContribution> getTasks()
    {
        checkInitialized();
        return Collections.unmodifiableSet(_taskContributions);
    }

    public TaskContribution getTaskByName( final String name )
    {
        checkInitialized();
        for (final TaskContribution taskContribution : _taskContributions)
        {
            if (taskContribution.getTaskName().equals(name))
            {
                return taskContribution;
            }
        }
        return null;
    }

    public void removeTaskExecutionContext(final UUID identifier)
    {
        checkInitialized();
        try
        {
            _taskLock.lock();

            if (!_availableTaskExecutionContexts.isEmpty())
            {
                for (TaskExecutionContext taskExecutionContext : _availableTaskExecutionContexts)
                {
                    if (taskExecutionContext.getIdentifier().equals(identifier))
                    {
                        _logger.info("Removing task execution context '" + identifier + "' from list of available task execution contexts");
                        _availableTaskExecutionContexts.remove(taskExecutionContext);
                        break;
                    }
                }

                // If no more tasks remain, raise an event stating that fact
                if (_availableTaskExecutionContexts.isEmpty())
                {
                    _logger.info("No more tasks available, raising event");
                    _eventManager.publish(new NoTasksAvailableEvent());
                }
            }

        }
        finally
        {
            _taskLock.unlock();
        }

    }

    private void checkInitialized()
    {
        if (_taskContributions == null)
        {
            throw new IllegalStateException("not initialized");
        }
    }

    public int getNumberOfTasks()
    {
        checkInitialized();
        return _taskContributions.size();
    }

    public void setModuleDescriptorTaskContributions(final List<ModuleDescriptorTaskContribution> moduleDescriptorTaskContributions)
    {
        if (moduleDescriptorTaskContributions == null)
        {
            throw new NullPointerException("moduleDescriptorTaskContributions");
        }
        _moduleDescriptorTaskContributions = moduleDescriptorTaskContributions;
    }

    public void initializeService()
    {
        if (_randomizer == null)
        {
            throw new IllegalStateException("no randomizer");
        }

        if (_logger == null)
        {
            throw new IllegalStateException("no log");
        }

        _taskContributions = new HashSet<TaskContribution>();

        _availableTaskExecutionContexts = new CopyOnWriteArrayList<TaskExecutionContext>();

        if (_taskProviders != null && _taskProviders.size() > 0)
        {
            for (TaskProvider taskProvider : _taskProviders)
            {
                _taskContributions.addAll(taskProvider.getTaskContributions());

                final List<Class> taskClasses = taskProvider.getTaskClasses();
                for (Class taskClass : taskClasses)
                {
                    final TaskConfiguration taskConfiguration = _taskConfigurationExtractor.extractTaskConfiguration( taskClass );
                    final Task task = _taskCreator.createTask(taskClass);
                    _taskContributions.add(new TaskContribution( task, taskConfiguration ));
                }
            }
        }
        else
        {
            _logger.debug("No task providers available");
        }

        if ( _moduleDescriptorTaskContributions != null && _moduleDescriptorTaskContributions.size() > 0)
        {
            for (ModuleDescriptorTaskContribution moduleDescriptorTaskContribution: _moduleDescriptorTaskContributions )
            {
                Task task = moduleDescriptorTaskContribution.getTask();
                if (task == null)
                {
                    Class taskClass = moduleDescriptorTaskContribution.getClazz();
                    if (taskClass != null)
                    {
                        task = _taskCreator.createTask(taskClass);
                    }
                    else
                    {
                        final Object instance = moduleDescriptorTaskContribution.getInstance();
                        if (instance != null)
                        {
                            task = _taskDecorator.decorateInstance(instance);
                        }
                    }
                }

                if (task == null)
                {
                    // No task instance found in the task info - deal with it
                    _logger.warn("No task instance could be created from the module task contribution: " + moduleDescriptorTaskContribution);
                    continue;
                }
                else
                {
                    _logger.debug("Task contribution added");
                    _taskContributions.add(new TaskContribution(task, moduleDescriptorTaskContribution));
                }
            }
            _logger.info("'" + _taskContributions.size() + "' task contribution(s) added");
        }
        else
        {
            _logger.warn("No task contributions supplied");
        }

    }

    public void basherEvent(final BasherEvent basherEvent)
    {
        if (basherEvent instanceof PhaseTransitionEvent)
        {
            PhaseTransitionEvent phaseTransitionEvent = (PhaseTransitionEvent) basherEvent;
            Phase newPhase = phaseTransitionEvent.getNewPhase();
            prepareTasksForPhase(newPhase, phaseTransitionEvent.getBasherContext());
        }
    }

    void prepareTasksForPhase(final Phase newPhase, final BasherContext basherContext)
    {
        _logger.info("Preparing available tasks for phase '" + newPhase + "'");
        _logger.info("Number available tasks: " + _taskContributions.size());

        // Clear existing lists
        _availableTaskExecutionContexts.clear();

        final List<TaskContribution> availableTasksWorkingList = new ArrayList<TaskContribution>();

        // First, filter the tasks for the phase in question
        filterForPhase(newPhase, availableTasksWorkingList);

        // Filter according to include/excludes defined in the BasherContext
        filterWithIncludesExcludes(basherContext, availableTasksWorkingList);

        // Sort them - in this case by task name
        sortTasks(availableTasksWorkingList);

        _logger.info("Preparation of tasks for phase '" + newPhase + "' completed");
        _logger.info("Number available tasks after preparation: " + availableTasksWorkingList.size());

        if (_logger.isDebugEnabled())
        {
            _logger.debug("Task contributions available after preparation:");
            for (TaskContribution task : availableTasksWorkingList)
            {
                _logger.debug("  " + task.getTaskName());
            }
        }

        // Finally, convert to a list of TaskExecutionContexts
        for (final TaskContribution taskContribution : availableTasksWorkingList)
        {
            _availableTaskExecutionContexts.add(new TaskExecutionContext(UUID.randomUUID(), taskContribution.getTask(), taskContribution));
        }


    }

    private void sortTasks(final List<TaskContribution> availableTasksWorkingList)
    {
        _logger.debug("Sorting tasks according to task name");
        // Sorting tasks
        Collections.sort(availableTasksWorkingList, new Comparator<TaskContribution>()
        {
            public int compare(final TaskContribution task1, final TaskContribution task2)
            {
                return task1.getTaskName().compareTo(task2.getTaskName());
            }
        });
    }

    private void filterWithIncludesExcludes(final BasherContext basherContext, final List<TaskContribution> availableTasksWorkingList)
    {
        _logger.debug("Filtering according to includes/excludes");
        // Filter tasks based on basher context include/exclude lists
        TaskTree taskTree = new TaskTree(availableTasksWorkingList);
        if (basherContext.getIncludes() != null)
        {
            _logger.debug("Applying include patterns: " + basherContext.getIncludes());
            taskTree.setIncludes(basherContext.getIncludes());
        }
        if (basherContext.getExcludes() != null)
        {
            _logger.debug("Applying exclude patterns: " + basherContext.getExcludes());
            taskTree.setExcludes(basherContext.getExcludes());
        }
        final List<TaskContribution> filteredTaskWorkingList = taskTree.getTasks();

        // Now, remove all that are not also part of the filtered list
        availableTasksWorkingList.retainAll(filteredTaskWorkingList);

        if (_logger.isDebugEnabled())
        {
            _logger.debug("Filtering according to includes/excludes completed");
            _logger.debug("Number available tasks after filtering: " + availableTasksWorkingList.size());
            _logger.debug("Tasks available after filtering: " + availableTasksWorkingList);
        }
    }

    private void filterForPhase(final Phase newPhase, final List<TaskContribution> availableTasksWorkingList)
    {
        _logger.debug("Filtering tasks for phase '" + newPhase + "'");
        // Add only tasks which apply to the phase in question
        for (TaskContribution taskContribution : _taskContributions)
        {
            if (taskContribution.getApplicablePhases().contains(newPhase))
            {
                availableTasksWorkingList.add(taskContribution);
            }
        }

        if (_logger.isDebugEnabled())
        {
            _logger.debug("Filtering for phase '" + newPhase + "' completed");
            _logger.debug("Number available tasks after filtering: " + availableTasksWorkingList.size());
            _logger.debug("Tasks available after filtering: " + availableTasksWorkingList);
        }
    }

}
