/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package net.sourceforge.basher.internal.impl;

import net.sourceforge.basher.*;
import net.sourceforge.basher.internal.Randomizer;
import net.sourceforge.basher.internal.TaskInvoker;
import net.sourceforge.basher.internal.TimeSource;
import org.apache.commons.logging.Log;

/**
 * @author Johan Lindquist
 * @version 1.0
 */
public class TaskInvokerImpl implements TaskInvoker
{
    private Randomizer _randomizer;
    private Collector _collector;
    private Log _logger;
    private TaskManager _taskManager;
    private TimeSource _timeSource;

    public void setTimeSource(final TimeSource timeSource)
    {
        _timeSource = timeSource;
    }

    public void setLog(final Log log)
    {
        _logger = log;
    }

    public void setRandomizer(final Randomizer randomizer)
    {
        _randomizer = randomizer;
    }

    public void setCollector(final Collector collector)
    {
        _collector = collector;
    }

    public void setTaskManager(final TaskManager taskManager)
    {
        _taskManager = taskManager;
    }

    public void invokeTask(final TaskExecutionContext taskExecutionContext )
    {
        doInvokeTask( taskExecutionContext, true);
    }

    public void invokeTask(final TaskExecutionContext taskExecutionContext, final boolean invokeFollowers)
    {
        doInvokeTask(taskExecutionContext, invokeFollowers);
    }

    private void doInvokeTask(final TaskExecutionContext taskExecutionContext, final boolean invokeFollower)
    {
        final Task task = taskExecutionContext.getTask();

        final TaskConfiguration taskConfiguration = taskExecutionContext.getTaskConfiguration();

        // First, reserve an invocation
        taskExecutionContext.reserveInvocation();

        try
        {
            // The number of invocations of the task - this is minus 1 since we have reserved one invocation
            final int invocations = taskExecutionContext.getInvocations() - 1;

            if (taskConfiguration.getMaxInvocations() != 0 && invocations >= taskConfiguration.getMaxInvocations())
            {
                if (_logger.isDebugEnabled())
                {
                    _logger.debug("Not running task: " + taskConfiguration.getTaskName() + " (max invocations reached)");
                }
                _collector.notRun(taskExecutionContext, 0, 0 );
                _taskManager.removeTaskExecutionContext(taskExecutionContext.getIdentifier());
                return;
            }

            final long currentTime = _timeSource.getElapsedTime() / 1000;

            if (currentTime < taskConfiguration.getRunFrom() || currentTime > taskConfiguration.getStopAfter())
            {
                if (_logger.isDebugEnabled())
                {
                    _logger.debug(String.format("Not running task: %s (not within time boundary %d/%d/%d)",taskConfiguration.getTaskName(),currentTime,taskConfiguration.getRunFrom(),taskConfiguration.getStopAfter()));
                }
                _collector.notRun(taskExecutionContext, 0, 0 );
                return;
            }

            final int weigher = _randomizer.getRandomInt(100);
            // We only run the task if the weigher is above the tasks threshold
            final int weight = taskConfiguration.getWeight();

            if (weigher < weight)
            {
                // Execute the task
                final long startTime = _timeSource.getCurrentTime();
                final long startTick = _timeSource.getNanoTime();
                try
                {
                    if (_logger.isDebugEnabled())
                    {
                        _logger.debug("Invoking task: " + taskConfiguration.getTaskName());
                    }
                    taskExecutionContext.executeTask();
                    // Should we invoke the subtasks?
                    if (invokeFollower && taskExecutionContext.getFollowers().size() != 0)
                    {
                        if (_logger.isDebugEnabled())
                        {
                            _logger.debug("Invoking followers");
                        }

                        for ( final TaskExecutionContext follower : taskExecutionContext.getFollowers() )
                        {
                            invokeTask( follower, false );
                        }
                    }

                    _collector.success(taskExecutionContext, _timeSource.getCurrentTime() - startTime, _timeSource.getNanoTime() - startTick);
                }
                catch (TaskNotRunException e)
                {
                    // collect it!
                    if (_logger.isDebugEnabled())
                    {
                        _logger.debug("Task not run: " + taskConfiguration.getTaskName() + " (" + e.getMessage() + ")");
                    }
                    _collector.notRun(taskExecutionContext, _timeSource.getCurrentTime() - startTime, _timeSource.getNanoTime() - startTick);
                }
                catch (Throwable throwable)
                {
                    // tell someone?
                    if (_logger.isErrorEnabled())
                    {
                        _logger.error("Task failed: " + taskConfiguration.getTaskName(), throwable);
                    }
                    _collector.fail(taskExecutionContext, _timeSource.getCurrentTime() - startTime, _timeSource.getNanoTime() - startTick, throwable);
                }
            }
            else
            {
                if (_logger.isDebugEnabled())
                {
                    _logger.debug(String.format("Not running task: %s (weight: %d / weigher: %d)", taskConfiguration.getTaskName(), weight, weigher));
                }
                _collector.notRun(taskExecutionContext, 0, 0 );
            }
        }
        finally
        {
            // and always release this invocation reservation
            taskExecutionContext.releaseInvocation();
        }
    }
}
