/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package net.sourceforge.basher.tasks;

import java.util.Collections;
import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;

import net.sourceforge.basher.Task;
import net.sourceforge.basher.TaskFailedException;
import net.sourceforge.basher.TaskNotRunException;
import org.apache.commons.logging.Log;

/**
 * A base implementation of the <code>Task</code> interface.  Provides some basic helper methods for dealing with
 * a task execution.
 *
 * @author Johan Lindquist
 * @version 1.0
 */
public abstract class AbstractTask implements Task
{
    /**
     * Instance logger
     */
    protected Log _log;

    /**
     * Convenience method for sub-classes to signal a not run.
     *
     * @param reason The reason for not running
     * @param cause  The (optional) cause of not run
     * @throws TaskNotRunException Will <b>ALWAYS</b> throw a <code>TaskNotRunException</code> to indicate the task didn't run.
     */
    protected void notRun(final String reason, final Throwable cause)
    {
        throw new TaskNotRunException(reason);
    }


    /**
     * Convenience method for sub-classes to signal a failure.
     *
     * @param reason The reason for the failure
     * @param cause  The (optional) cause of the failure
     * @throws TaskFailedException Will <b>ALWAYS</b> throw a <code>TaskFailedException</code> to indicate a failure.
     */
    protected void failed(final String reason, final Throwable cause)
    {
        throw new TaskFailedException(reason, cause);
    }

    /**
     * Convenience method for sub-classes to signal a failure.
     *
     * @param reason The reason for the failure
     * @throws TaskFailedException Will <b>ALWAYS</b> throw a <code>TaskFailedException</code> to indicate a failure.
     */
    protected void failed(final String reason)
    {
        failed(reason, null);
    }

    /**
     * Sets the <code>Log</code> instance for this <code>Task</code> instance.
     *
     * @param log The log instance to use for logging.
     */
    public void setLog(final Log log)
    {
        _log = log;
    }

    @Override
    public String toString()
    {
        return "AbstractTask{" +
                '}';
    }

    @Override
    public boolean equals( final Object o )
    {
        if ( this == o )
        {
            return true;
        }
        if ( !( o instanceof AbstractTask ) )
        {
            return false;
        }

        final AbstractTask that = ( AbstractTask ) o;

        if ( _log != null ? !_log.equals( that._log ) : that._log != null )
        {
            return false;
        }

        return true;
    }

    @Override
    public int hashCode()
    {
        int result = _log != null ? _log.hashCode() : 0;
        return result;
    }
}
